from .__version__ import __version__
from .agmri_api import AgmriApi
from .alert_service_api import AlertServiceApi
from .annotations_api import AnnotationsApi
from .logging import *
from .ml_api import MLApi
from .s3_utils import *
from .os_utils import *
from .efs_utils import *
from .ssm_utils import *



"""
This library provides an interface to the REST API exposed by Ag-MRI.

This library acts as a thin authentication library for the Requests library.
When performing any requests that library documentation should be referenced.

Admin credentials are pulled from the ~/.agmri.cfg file. This file is in the
.ini file format, which is documented here: https://docs.python.org/3.5/library/configparser.html
It should contain a section for each environment with an admin_username and
admin_password value within it.

Example:
```ini
[release]
admin_username = EMAIL
admin_password = PASSWORD

[prod]
admin_username = EMAIL
admin_password = PASSWORD
```

The section headers should match the value that is specified when
initializing the Api() class. An example:

```python
from commons.agmri_api import Api

api = Api("prod")  # This will work with the above cfg file
api = Api("preprod")  # So will this
api = Api("testing")  # This will fail because it's missing from the cfg file
```

"""

import configparser
import logging
import os
from typing import Union

import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import Timeout, ConnectionError, ChunkedEncodingError, HTTPError
from urllib.parse import urlparse
from urllib3.util.retry import Retry
from string import Template
import boto3
import deprecation

import intelinair_utils.token_store as token_stores
from intelinair_utils.ssm_utils import get_ssm_parameter
from intelinair_utils.api import Api
from intelinair_utils.api_utils import validate_date

URL_TEMPLATE = "https://api.{}.intelinair.dev/admin/api/"
REQUEST_TIMEOUT = 180
RETRIES = 7

logger = logging.getLogger(__name__)


class AgmriApi(Api):
    """
    Wrapper class for authenticated queries against Ag-MRI

    After initialization, authentication has been completed and requests
    through this object will be automatically authenticated. Only GET,
    POST and PATCH are supported right now.
    """

    def __init__(self, environment: str, config_path: str = None, version_string: str = 'v2/'):
        # version_string == '' corresponds to the first version of API
        self.environment = environment
        if environment == 'prod':
            self.api_url = "https://api.ag-mri.intelinair.com/admin/api/"
        else:
            self.api_url = URL_TEMPLATE.format(environment)
        self._config_path = config_path
        # all calls (other than login) should use the version string
        assert version_string.endswith('/')
        self.api_version_url = self.api_url + version_string
        self.url_template = f"{self.api_version_url}{{}}"
        self.session = requests.Session()
        retries = Retry(total=10, read=10, connect=10, backoff_factor=1,
                        allowed_methods=False,
                        status_forcelist=[409] + list(range(500, 600)))
        self.tokens = {}
        self.headers = {
            'processing': {'X-Processing': 'true'},
            'graphql': {'Content-Type': 'application/graphql'},
            'refresh': {'Content-Type': 'application/x-www-form-urlencoded'}
        }
        self.session.mount('https://', HTTPAdapter(max_retries=retries))
        self._config = None
        self._token_store = None

    @property
    def config(self):
        if self._config is None:
            self._config = configparser.ConfigParser()
            self._config['DEFAULT']['token_store'] = 'FileStore'

            if self._config_path is not None:
                parsed_path = urlparse(self._config_path)
                if parsed_path.scheme == "file":
                    self._config.read(parsed_path.path)
                elif parsed_path.scheme == "ssm":
                    ssm_client = boto3.client('ssm')
                    # get content from the ssm
                    config_content = get_ssm_parameter(ssm_client, parsed_path.path)
                    self._config.read_string(config_content)
                else:
                    raise Exception("Not supported configuration format. Please use file:// or ssm://")
            else:
                # checking the env variable for config content
                if "AGMRI_API_CONFIG" in os.environ:
                    self._config.read_string(os.environ['AGMRI_API_CONFIG'])
                else:
                    self._config.read(os.path.expanduser("~/.agmri.cfg"))

            if not self._config.has_section(self.environment):
                raise Exception("Missing settings for {} environment in".format(self.environment))
        return self._config

    @property
    def user(self):
        return {
            'username': self.config.get(self.environment, 'admin_username'),
            'password': self.config.get(self.environment, 'admin_password')
        }

    @property
    def token_store(self):
        if self._token_store is None:
            driver = self.config.get(self.environment, 'token_store')
            self._token_store = getattr(token_stores, driver)(self.environment, self.user['username'])
        return self._token_store

    def login(self):

        # login URL does not use version url. it's always /api/login
        logger.info(f'Logging in with to {self.api_url} with username={self.user["username"]}')
        auth_response = self.make_request(self.session.post, self.api_url + "login", json=self.user, ignore_json=False, attempt_authentication=False)
        if auth_response is not None:
            self.tokens['access_token'] = auth_response['access_token']
            self.tokens['refresh_token'] = auth_response['refresh_token']
            self.headers['processing']['X-Auth-Token'] = auth_response['access_token']
        else:
            raise Exception("Failed to login into ag-mri")

    def token_is_valid(self):
        if 'access_token' not in self.tokens:
            return False

        url = os.path.join(self.api_url, "validate")
        try:
            res = self.make_request(
                self.session.post,
                url,
                headers={'X-Auth-Token': self.tokens['access_token']},
                ignore_json=True,
                attempt_authentication=False
            )
        except HTTPError as e:
            logger.debug(f'token not valid with exception {e}')
            return False
        else:
            if res.status_code == 200:
                return True
            else:
                return False

    def refresh_token(self):
        url = self.api_url.replace("api/", "oauth/access_token")
        params = {
            'grant_type': 'refresh_token',
            'refresh_token': self.tokens['refresh_token']
        }
        res = self.make_request(
            self.session.post,
            url,
            params=params,
            headers=self.headers['refresh'],
            attempt_authentication=False
        )
        self.tokens['access_token'] = res['access_token']
        self.tokens['refresh_token'] = res['refresh_token']
        self.headers['processing']['X-Auth-Token'] = res['access_token']

    def ensure_authenticated(self):
        if self.token_is_valid():
            return

        try:
            self.tokens = self.token_store.get_tokens()
            if self.token_is_valid():
                self.headers['processing']['X-Auth-Token'] = self.tokens['access_token']
                return
        except Exception as e:
            logger.debug(f'failed to fetch tokens with exception {e}')

        try:
            self.refresh_token()
        except Exception as e:
            logger.debug(f'failed to refresh tokens with exception {e}')
        else:
            self.token_store.store_tokens(self.tokens)
            return

        self.login()
        self.token_store.store_tokens(self.tokens)

    def make_request(self, http_method, *args, ignore_json=False, attempt_authentication=True, **kwargs):

        response = http_method(*args, **kwargs, timeout=REQUEST_TIMEOUT)
        status_code = response.status_code

        if status_code == 401 and attempt_authentication:
            self.ensure_authenticated()
            kwargs['headers'].update(self.headers['processing'])
            return self.make_request(http_method, *args, ignore_json=ignore_json, attempt_authentication=False, **kwargs)

        response.raise_for_status()

        if status_code == 204 or ignore_json:
            return response

        return response.json()

    @deprecation.deprecated(details="Please use specific function instead of direct end point request.")
    def get(self, path, params=None, ignore_json=False, **kwargs):
        """Perform a GET request to the specified endpoint"""
        return self.make_request(self.session.get, self.api_version_url + path.lstrip('/'), headers=self.headers['processing'], params=params, ignore_json=ignore_json, **kwargs)

    @deprecation.deprecated(details="Please use specific function instead of direct end point request.")
    def post(self, path, params=None, ignore_json=False, **kwargs):
        # basically this extends http://docs.python-requests.org/en/latest/api/#requests.request
        # in particular, kwargs can contain files array
        """Perform a POST request to the specified endpoint"""
        return self.make_request(self.session.post, self.api_version_url + path.lstrip('/'), headers=self.headers['processing'], params=params, ignore_json=ignore_json, **kwargs)

    @deprecation.deprecated(details="Please use specific function instead of direct end point request.")
    def delete(self, path, params=None, ignore_json=True):
        """Perform a POST request to the specified endpoint"""
        return self.make_request(self.session.delete, self.api_version_url + path.lstrip('/'), headers=self.headers['processing'], params=params, ignore_json=ignore_json)

    @deprecation.deprecated(details="Please use specific function instead of direct end point request.")
    def post_json(self, path, json=None, ignore_json=False):
        """Perform a POST request to the specified endpoint"""
        return self.make_request(self.session.post, self.api_version_url + path.lstrip('/'), headers=self.headers['processing'], json=json, ignore_json=ignore_json)

    @deprecation.deprecated(details="Please use specific function instead of direct end point request.")
    def patch_json(self, path, json=None, ignore_json=True):
        """Perform a PATCH request to the specified endpoint"""
        return self.make_request(self.session.patch, self.api_version_url + path.lstrip('/'), headers=self.headers['processing'], json=json, ignore_json=ignore_json)

    @deprecation.deprecated(details="Please use specific function instead of direct end point request.")
    def put_json(self, path, json=None, ignore_json=False):
        # TODO set ignore_json=True in after making sure that json problem is not connected with response code
        """Perform a PUT request to the specified endpoint"""
        return self.make_request(self.session.put, self.api_version_url + path.lstrip('/'), headers=self.headers['processing'], json=json, ignore_json=ignore_json)

    def graphql_request(self, service, query):

        if 'X-Auth-Token' not in self.headers['processing']:
            self.ensure_authenticated()

        headers = {**self.headers['processing'], **self.headers['graphql']}
        if self.environment == 'prod':
            host = 'api.ag-mri.intelinair.com'
        else:
            host = 'api.{}.intelinair.dev'.format(self.environment)

        graphql_url = "https://{}/{}/graphql".format(host, service)

        response = self.make_request(self.session.post, graphql_url, headers=headers, data=query)
        return response

    def fields_report(self, request_fields, extent=None, only_monitored=True, enabled=True, company_deleted=False,
                      prospecting_mode=None, only_in_grid=None, satellite_only=None, company_id=None, provider=None,
                      flight_date=None, r_max=10000, offset=0):

        query = Template("""{
          fieldsReport(queryCommand: {
            $params
          }, max: $r_max, offset: $offset) {
                totalCount
                results {
                    $results
                }
            }
        }""")

        params = ""
        if extent is not None:
            params += """northEastLatitude: {},
            northEastLongitude: {},
            southWestLatitude: {},
            southWestLongitude: {},""".format(extent[0], extent[1], extent[2], extent[3])
        if only_monitored is True:
            params += """
            onlyMonitored: true,"""
        if enabled is True:
            params += """
            enabled: true,"""
        if company_deleted is False:
            params += """
            companyDeleted: false,"""
        if only_in_grid is True:
            params += """
            onlyInGrid: true,"""
        if satellite_only:
            params += """
            satelliteOnly: true,"""
        if prospecting_mode:
            params += """
            prospectingMode: {},""".format(prospecting_mode)
        if company_id:
            params += """
            companyId: {},""".format(company_id)
        if provider is not None:
            params += """
            provider: \"{}\",""".format(provider)
        if flight_date is not None:
            params += """
            flightDate: \"{}\",""".format(flight_date)
        params += """
            sort: \"id\""""

        results = ""
        if 'id' in request_fields:
            results += "id"
        if 'name' in request_fields:
            results += """
                    name"""
        if 'area' in request_fields:
            results += """
                    area"""
        if 'tags' in request_fields:
            results += """
                    tags"""
        if 'description' in request_fields:
            results += """
                    description"""
        if 'grid_id' in request_fields:
            results += """
                    grid{
                        id
                    }"""
        if 'prospecting_mode' in request_fields:
            results += """
                    farm{
                        division{
                            company{
                                alias{
                                    companyHierarchy{
                                      prospectingMode
                                    }
                                }
                            }
                        }
                    }"""
        if 'latestFlight' in request_fields:
            results += """
                    latestFlight{
                        code
                    }"""
        if 'geometry' in request_fields:
            results += """
                    geoData{
                        geometry
                    }"""

        query = query.substitute(params=params, results=results, r_max=r_max, offset=offset)
        data = self.graphql_request(service="reports", query=query)
        return data

    def get_field_info(self, field_id, request_fields, service='admin'):
        """get field info

        Args:
            field_id: field id
            request_fields: fields to request available fields are
                id, name, latestFlight, fieldSeason, cropTypes, plantingDate, tags, area, token, plantedSeeds, flights
            service: service to make a call. Defaults to 'admin'.

        Returns:
            dict: field info
        """        
        query = Template(
        """{
          field(id: $field_id) {
            $params
            } 
        }""")
        params = ""

        if 'id' in request_fields:
            params += """
            id"""
        if 'name' in request_fields:
            params += """
            name"""
        if 'latestFlight' in request_fields:
            params += """
            latestFlight{
                id
                date
            }"""
        if 'fieldSeason' in request_fields:
            params += """
            fieldSeason{
                id
                geoData{
                    id
                    geometry
                }
            }"""
        if 'cropTypes' in request_fields:
            params += """
            cropTypes {
                id
                name
            }"""
        if 'plantingDate' in request_fields:
            params += """
            plantingDate"""
        if 'tags' in request_fields:
            params += """
            tags"""
        if 'area' in request_fields:
            params += """
            area"""
        if 'token' in request_fields:
            params += """
            token"""
        if 'plantedSeeds' in request_fields:
            params += """
            plantedSeeds"""
        if 'flights' in request_fields:
            params += """
            flights {
                id,
                date,
                code
                season {
                    id
                }
                provider
                reviewStatus
            }"""
        query = query.substitute(params=params, field_id=field_id)
        data = self.graphql_request(service=service, query=query)
        return data.get('data', {}).get('field', {})

    def get_attachment_types(self):
        types = self.get("fieldattachmenttypes?max=100")
        return {z['keyName']: z['id'] for z in types['fieldAttachmentTypes']}

    # GET requests
    def get_flight(self, flight_code: str) -> dict:
        """
        Getting flight object.
        Args:
            flight_code:
        """
        url = self.url_template.format(f"flights/{flight_code}")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_imagery(self, imagery_code: str) -> dict:
        """
        Getting imagery data.
        Args:
            imagery_code:
        """
        url = self.url_template.format(f"imagery/{imagery_code}")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_seasons(self) -> dict:
        """
        Getting season info.
        Returns:
            list of seasons info wrapped in dict.
        """
        url = self.url_template.format("season/")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_field(self, field_id: Union[int, str]) -> dict:
        """
        getting field info for a given field_id
        Args:
            field_id:
        """
        url = self.url_template.format(f"fields/{field_id}")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_flight_processing_data(self, flight_code: str) -> dict:
        """
        Getting processing data for the flight.
        Args:
            flight_code:
        """
        url = self.url_template.format(f"flights/{flight_code}/processing")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_equipment_flight_data(self, flight_code: str) -> dict:
        """
        Getting equipment flight data.
        Args:
            flight_code:
        """
        url = self.url_template.format(f"flights/{flight_code}/equipmentData")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_uploaded_files(self, file_id: Union[int, str], params: dict) -> dict:
        """
        Getting uploaded file by ID.
        Args:
            file_id:
            params: http request params as key values.
        """
        url = self.url_template.format(f"uploadFile/{file_id}/")
        response = self.make_request(self.session.get, url, params=params, headers=self.headers['processing'])
        return response

    def get_field_attachments(self, field_id: Union[int, str]) -> list:
        """
        Getting field attachements list.
        Args:
            field_id:
        """
        url = self.url_template.format(f"fields/{field_id}/attachments")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_field_processing_data(self, field_id: Union[int, str]) -> list:
        """
        Get processing data for the field.
        Args:
            field_id:
        """
        url = self.url_template.format(f"fields/{field_id}/processing")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_field_all_boundaries(self, field_id: Union[int, str], ignore_json=True) -> dict:
        """
        Getting field all boundaries.
        Args:
            field_id:
            ignore_json:
        """
        url = self.url_template.format(f"fields/{field_id}/allboundaries")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'], ignore_json=ignore_json)
        return response

    def get_previous_flight(self, flight_code: str) -> dict:
        """
        Getting previous flight of the given one.
        Args:
            flight_code:
        """
        url = self.url_template.format(f"flights/{flight_code}/previous")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_field_tilesets(self, field_id: Union[int, str]) -> dict:
        """
        Getting list if tilesets for a given field ID.
        Args:
            field_id:
        """
        url = self.url_template.format(f"fields/{field_id}/tilesets")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def refresh_field_report(self):
        """
        Refresh field report.
        """
        url = self.url_template.format("fields/refreshFieldReport")
        self.make_request(self.session.get, url, headers=self.headers['processing'])

    def get_flight_alerts(self, flight_code: str, alert_name: str) -> dict:
        """
        Getting flight alerts for the given alert name and flight code.
        Args:
            flight_code:
            alert_name:
        """
        url = self.url_template.format(f"flights/{flight_code}/alerts?kindKeys={alert_name}")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_imageries(self, filters: dict) -> dict:
        """
        Getting imageries filtered by given constraints in filters.
        Args:
            filters: key values based on which output will be generated. e.g.
            {
                "startDate": 2022-05-01,
                "endDate": 2022-05-31,
                "seasonId": 71,
                "provider": "Sentinel2"
            }
        """
        url = self.url_template.format("imagery/")
        response = self.make_request(self.session.get, url, params=filters, headers=self.headers['processing'])
        return response

    def get_growing_degree_days(self, flight_code: str, date_from: str, date_to: str) -> dict:
        """
        Get GDD for the flight.
        Args:
            flight_code:
            date_from:
            date_to:
        """
        validate_date(date_from)
        validate_date(date_to)
        params = {
            'flightCode': flight_code,
            'dateFrom': date_from,
            'dateTo': date_to
        }
        url = self.url_template.format("flights/gddForFlight")
        response = self.make_request(self.session.get, url, params=params, headers=self.headers['processing'])
        return response

    def get_last_frost_date(self, field_id: Union[int, str], season_id: int, date_to: str) -> dict:
        """
        Get last frost date for the field.
        Args:
            field_id:
            season_id:
            date_to:
        """
        validate_date(date_to)
        url = self.url_template.format(f'fields/{field_id}/lastFrost')
        params = {
            'seasonId': season_id,
            'dateTo': date_to
        }
        response = self.make_request(self.session.get, url, params=params, headers=self.headers['processing'])
        return response

    def get_rollback_gdd_date(self, field_id: Union[int, str], season_id: int, date_from: str, rollback=100) -> dict:
        """
        Get rollback gdd date for the field.
        Args:
            field_id:
            season_id:
            date_from:
            rollback:
        """
        validate_date(date_from)
        url = self.url_template.format(f'fields/{field_id}/rollbackGdd')
        params = {
            'seasonId': season_id,
            'dateFrom': date_from,
            'delta': rollback
        }
        response = self.make_request(self.session.get, url, params=params, headers=self.headers['processing'])
        return response

    def get_field_attachment_types(self, max_count=100):
        """
        Getting available attachment types for the field.
        Args:
            max_count: Maximum count of types to return.
        """
        url = self.url_template.format("fieldattachmenttypes")
        params = {
            'max': max_count
        }
        response = self.make_request(self.session.get, url, params=params, headers=self.headers['processing'])
        return response

    def check_field_weather_data_existence(self, field_id: Union[int, str], start_date: str, end_date: str) -> dict:
        """
        Checking given field weather data download completion status.
        Args:
            field_id:
            start_date:
            end_date:
        """
        validate_date(start_date)
        validate_date(end_date)
        url = self.url_template.format("weather/verifyConsistency")
        params = {
            'fieldId': field_id,
            'startDate': start_date,
            'endDate': end_date
        }
        response = self.make_request(self.session.get, url, params=params, headers=self.headers['processing'])
        return response

    def get_company(self, company_id: Union[int, str]) -> dict:
        """
        Getting company info from given ID.
        Args:
            company_id:
        """
        url = self.url_template.format(f"companies/{company_id}")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_flight_tilesets(self, flight_code: str) -> dict:
        """
        Getting flight tilesets.
        Args:
            flight_code:
        """
        url = self.url_template.format(f"flights/{flight_code}/tilesets")
        response = self.make_request(self.session.get, url, headers=self.headers['processing'])
        return response

    def get_problem_alert_default_tileset(self, flight_code: str, badge_id: Union[int, str]) -> dict:
        """
        Getting default tileset for problem alert
        Args:
            flight_code:
            badge_id: problem alert badge ID
        """
        params = {
            'badgeId': badge_id
        }
        url = self.url_template.format(f"flights/{flight_code}/defaultTileset")
        response = self.make_request(self.session.get, url, params=params, headers=self.headers['processing'])
        return response

    # POST requests
    def imageries_report(self, filters: dict) -> dict:
        """
        Getting reported imagery list based on filters.
        Args:
            filters: key values based on which output will be generated. e.g.
            {
                "startDate": 2022-05-01,
                "endDate": 2022-05-31
            }
        """
        url = self.url_template.format("reports/imagery/")
        response = self.make_request(self.session.post, url, params=filters, headers=self.headers['processing'])
        return response

    def upload_complete_flight(self, flight_code: str):
        """
        Setting upload Complete to the flight.
        Args:
            flight_code:
        """
        url = self.url_template.format(f"flights/{flight_code}/uploadComplete")
        self.make_request(self.session.post, url, headers=self.headers['processing'])

    def get_field_equipment_flight_data(self, field_id: Union[int, str], season_id: int) -> dict:
        """
        Getting list of equipment flights data for the field.
        Args:
            field_id:
            season_id:
        """
        url = self.url_template.format("flights/equipmentFlightsData")
        params = {
            "fieldId": field_id,
            "seasonId": season_id
        }
        response = self.make_request(self.session.post, url, params=params, headers=self.headers['processing'])
        return response

    def upload_field_attachment(self, field_id: Union[int, str], attachment_id: Union[int, str], attachment_path: str):
        """
        Uploading field attachment file.
        Attachment is done through binary stream.
        Args:
            field_id:
            attachment_id:
            attachment_path:
        """
        url = self.url_template.format(f"fields/{field_id}/attachments/{attachment_id}/upload")
        with open(attachment_path, 'rb') as attachment_handle:
            files = {
                (os.path.basename(attachment_path),
                 attachment_handle,
                 'application/octet-stream')
            }
            self.make_request(self.session.post, url, files=files, headers=self.headers['processing'])

    def rerun_flight(self, flight_code: str):
        """
        Rerunning given flight processing.
        Args:
            flight_code:
        """
        url = self.url_template.format(f"flights/{flight_code}/rerun")
        self.make_request(self.session.post, url, headers=self.headers['processing'])

    # DELETE requests
    def delete_flight(self, flight_code: str):
        """
        Delete flight by code.
        Args:
            flight_code:
        """
        url = self.url_template.format(f"flights/{flight_code}")
        self.make_request(self.session.delete, url, headers=self.headers['processing'], ignore_json=True)

    def delete_field_attachment(self, field_id: Union[int, str], attachment_id: Union[int, str]):
        """
        Delete field attachment file by given field and file id.
        Args:
            field_id:
            attachment_id:
        """
        url = self.url_template.format(f"fields/{field_id}/attachments/{attachment_id}")
        self.make_request(self.session.delete, url, headers=self.headers['processing'], ignore_json=True)

    def delete_field_tileset(self, field_id: Union[int, str], tileset_id: Union[int, str]):
        """
        Delete field tileset by id.
        Args:
            field_id:
            tileset_id:
        """
        url = self.url_template.format(f"fields/{field_id}/tilesets/{tileset_id}")
        self.make_request(self.session.delete, url, headers=self.headers['processing'], ignore_json=True)

    # POST_JSON requests
    def get_flights_annotations(self, filters: dict, params=None) -> dict:
        """
        Getting flight annotations report based on input params in filter.
        Args:
            filters: key values based on which output will be generated. e.g.
            {
                "alertKindId": 29,
                "annotationStatus": "QA_COMPLETE"
            }
            params: http request params as key values.
        """
        url = self.url_template.format("reports/toafa")
        response = self.make_request(self.session.post, url, filters, params=params, headers=self.headers['processing'])
        return response

    def add_field_tag(self, field_id: Union[int, str], tag: str, key_name='dashboard'):
        """
        Add tag to field object.
        Args:
            field_id:
            tag: Tag name to add.
            key_name: tag keyname (mainly dashboard and system).
        """
        url = self.url_template.format(f"fields/{field_id}/addTag")
        params = {
            "optionName": tag,
            "keyName": key_name
        }
        self.make_request(self.session.post, url, json=params, headers=self.headers['processing'])

    def remove_field_tag(self, field_id: Union[int, str], tag: str, key_name='dashboard'):
        """
        Remove given tag from the field object.
        Args:
            field_id:
            tag: tag name.
            key_name: tag keyname (mainly dashboard and system).
        """
        url = self.url_template.format(f"fields/{field_id}/removeTag")
        params = {
            "optionName": tag,
            "keyName": key_name
        }
        self.make_request(self.session.post, url, json=params, headers=self.headers['processing'])

    def add_flight_tag(self, flight_code: str, tag: str, key_name='dashboard'):
        """
        Add tag to the flight object.
        Args:
            flight_code:
            tag: tag name to add.
            key_name: tag keyname (mainly dashboard and system).
        """
        url = self.url_template.format(f"flights/{flight_code}/addTag")
        params = {
            "optionName": tag,
            "keyName": key_name
        }
        self.make_request(self.session.post, url, json=params, headers=self.headers['processing'])

    def remove_flight_tag(self, flight_code: str, tag: str, key_name='dashboard'):
        """
        Remove given tag from the flight object.
        Args:
            flight_code:
            tag: tag name.
            key_name: tag keyname (mainly dashboard and system).
        """
        url = self.url_template.format(f"flights/{flight_code}/removeTag")
        params = {
            "optionName": tag,
            "keyName": key_name
        }
        self.make_request(self.session.post, url, json=params, headers=self.headers['processing'])

    def get_field_flight(self, field_id: Union[int, str], filters: dict) -> dict:
        """
        Getting flight for a given field based on input filters.
        Args:
            field_id:
            filters: key values based on which output will be generated. e.g.
            {
                'provider': 'Aeroptic',
                'date': 2022-05-01,
            }
        """
        url = self.url_template.format(f"fields/{field_id}/flight")
        response = self.make_request(self.session.post, url, json=filters, headers=self.headers['processing'])
        return response

    def flights_report(self, filters: dict) -> dict:
        """
        Generates report of the flights based on a given filters.
        Args:
            filters: key values based on which output will be generated. e.g.
            {
                'fieldId': 12345678,
                'providers': ['Aeroptic'],
                'startDate': 2022-05-01,
                'endDate': 2022-05-30
            }
        """
        url = self.url_template.format("reports/toaf")
        response = self.make_request(self.session.post, url, json=filters, headers=self.headers['processing'])
        return response

    def create_imagery(self, payload: dict) -> dict:
        """
        Creates imagery object with values passed through payload.
        Args:
            payload: properties of imagery object to create.
        Returns:
            Created imagery object.
        """
        url = self.url_template.format("imagery/")
        response = self.make_request(self.session.post, url, json=payload, headers=self.headers['processing'])
        return response

    def create_flight(self, payload: dict) -> dict:
        """
        Creates flight object with values passed through payload.
        Args:
            payload: properties of flight object to create.
        Returns:
            Created flight object.
        """
        url = self.url_template.format("flights/")
        response = self.make_request(self.session.post, url, json=payload, headers=self.headers['processing'])
        return response

    def create_imagery_tile_layer(self, payload: dict):
        """
        Creates imagery layer.
        Args:
            payload: properties of tile layer to create.
        """
        url = self.url_template.format("imagery/tiles/layers")
        self.make_request(self.session.post, url, json=payload, headers=self.headers['processing'])

    def post_field_metadata(self, field_id: Union[int, str], metadata: dict) -> dict:
        """
        Create field metadata.
        Args:
            field_id:
            metadata:
        """
        url = self.url_template.format(f"fields/{field_id}/metadata/")
        response = self.make_request(self.session.post, url, json=metadata, headers=self.headers['processing'],
                                     ignore_json=True)
        return response

    def toggle_flight_tileset(self, flight_code: str, tileset_name: str, is_hidden: bool):
        """
        Changing hidden attribute of the flight tileset.
        Args:
            flight_code: on which change should be performed.
            tileset_name: tileset which will be changed.
            is_hidden: value to be set.
        """
        url = self.url_template.format(f"flights/{flight_code}/toggleTilesets/")
        payload = {
            'tilesets': [
                {
                    'name': tileset_name,
                    'hidden': is_hidden
                }
            ]
        }
        self.make_request(self.session.post, url, json=payload, headers=self.headers['processing'],
                          ignore_json=True)

    def reject_flight(self, flight_code: str, comment: str):
        """
        Automatically reject the flight with a given comment.
        Args:
            flight_code:
            comment: comment of rejection.
        """
        url = self.url_template.format(f"flights/{flight_code}/flightComments/")
        payload = {
            "reviewer": "Automatic",
            "flightCode": flight_code,
            "subject": "Automatic rejection",
            "comments": comment
        }
        self.make_request(self.session.post, url, json=payload, headers=self.headers['processing'])

    def create_field_attachment(self, field_id: Union[int, str], attachment_data: dict):
        """
        Create field attachment with given data.
        Args:
            field_id:
            attachment_data:
        """
        url = self.url_template.format(f'/fields/{field_id}/attachments')
        self.make_request(self.session.post, url, json=attachment_data, headers=self.headers['processing'])

    def create_field_tilesets(self, field_id: Union[int, str], tilesets_data: dict):
        """
        Creates tilesets from data for the given field.
        Args:
            field_id:
            tilesets_data:
        """
        url = self.url_template.format(f'/fields/{field_id}/tilesets')
        self.make_request(self.session.post, url, json=tilesets_data, headers=self.headers['processing'])

    def create_flight_tilesets(self, flight_code: str, tilesets_data: dict):
        """
        Creates tilesets from data for the given flight.
        Args:
            flight_code:
            tilesets_data:
        """
        url = self.url_template.format(f'/flights/{flight_code}/tilesets')
        self.make_request(self.session.post, url, json=tilesets_data, headers=self.headers['processing'])

    # PATCH_JSON requests
    def update_imagery(self, imagery_code: str, imagery_data: dict):
        """
        Update imagery object properties.
        Args:
            imagery_code: code of imagery to be changed.
            imagery_data: key-values of properties that will be changed.
        """
        url = self.url_template.format(f"imagery/{imagery_code}/")
        self.make_request(self.session.patch, url, json=imagery_data, headers=self.headers['processing'],
                          ignore_json=True)

    def update_field(self, field_id: Union[int, str], field_data: dict):
        """
        Update field object properties.
        Args:
            field_id: id of field to be changed.
            field_data: key-values of properties that will be changed.
        """
        url = self.url_template.format(f"fields/{field_id}/")
        self.make_request(self.session.patch, url, json=field_data, headers=self.headers['processing'],
                          ignore_json=True)

    def update_field_attachment(self, field_id: Union[int, str], attachment_id: Union[int, str], attachment_data: dict):
        """
        Update field attachment file
        Args:
            field_id:
            attachment_id: id of a file to be changed.
            attachment_data: key-values of properties that will be changed.
        """
        url = self.url_template.format(f"fields/{field_id}/attachments/{attachment_id}/")
        self.make_request(self.session.patch, url, json=attachment_data, headers=self.headers['processing'],
                          ignore_json=True)

    def update_flight(self, flight_code: str, flight_data: dict):
        """
        Update flight object.
        Args:
            flight_code: code of flight to change.
            flight_data: key-values of properties that will be changed.
        """
        url = self.url_template.format(f"flights/{flight_code}/")
        self.make_request(self.session.patch, url, json=flight_data, headers=self.headers['processing'],
                          ignore_json=True)

    def update_flight_tilesets(self, flight_id: Union[int, str], tileset_id: Union[int, str], tilesets_data: dict):
        """
        Update given flight tilesets.
        Args:
            flight_id:
            tileset_id:
            tilesets_data: updated data
        """
        url = self.url_template.format(f"flights/{flight_id}/tilesets/{tileset_id}")
        self.make_request(self.session.patch, url, json=tilesets_data, headers=self.headers['processing'],
                          ignore_json=True)

    # PUT_JSON requests
    def update_uploaded_file(self, file_id: Union[int, str], file_data: dict):
        """
        Update uploaded file.
        Args:
            file_id: id of file to change.
            file_data: key-values of properties that will be changed.
        """
        url = self.url_template.format(f"uploadFile/{file_id}/")
        self.make_request(self.session.put, url, json=file_data, headers=self.headers['processing'])

    def update_flight_alert(self, flight_code: str, alert_id: Union[int, str], alert_data: dict):
        """
        Update alert of the flight.
        Args:
            flight_code: code of a flight to change.
            alert_id: id of an alert to change.
            alert_data: key-values of properties that will be changed.
        """
        url = self.url_template.format(f"flights/{flight_code}/alerts/{alert_id}/")
        self.make_request(self.session.put, url, json=alert_data, headers=self.headers['processing'])

    def add_field_list_to_file(self, file_id: Union[int, str], field_list: list):
        """
        add field list to the uploaded file.
        Args:
            file_id: uploaded file id
            field_list: list of fields to be added
        """
        url = self.url_template.format(f"uploadFile/{file_id}/addToFieldList")
        payload = {
            "fields": field_list
        }
        self.make_request(self.session.put, url, json=payload, headers=self.headers['processing'])



import logging
import os
import sys

__all__ = ['set_standard_logging_config']


def suppress_common_packages():
    """Suppress INFO logs from common packages to reduce clutter"""
    # setting warning level to various packages to avoid additional logs
    packages = ('botocore', 'boto3', 'requests', 'urllib3', 'rasterio', 'shapely', 'paramiko')
    for p in packages:
        p_logger = logging.getLogger(p)
        p_logger.setLevel(logging.WARNING)


def get_fmt_string(code: str = None, env: str = None) -> str:
    """Creates common format string for logging

    Args:
        code: an optional flight code to include
        env: an optional environment to include

    Returns:
        a format string
    """
    format_str = '[{asctime}][{name}][{filename}:{lineno}][{funcName}][{levelname}]: {message}'

    if code:
        format_str = "[{}]{}".format(code, format_str)

    if env:
        format_str = "[{}]{}".format(env, format_str)

    return format_str


def set_standard_logging_config(*, level=logging.INFO, stdout: bool = True, stderr: bool = False, logfile: str = None,
                                code: str = None, env: str = None, suppress_packages: bool = True):
    """Sets a standard logging context

    Args:
        level: the logging level of the root logger
        stdout: if true will print logs to stdout
        stderr: if true will print logs to stderr
        logfile: an optional path to write logs to
        code: an optional code to include in the logging output
        env: an optional environment to include in the logging output
        suppress_packages: if true will suppress info messages from common packages
    """

    if env is None and 'IA_ENV' in os.environ:
        env = os.environ['IA_ENV']

    if suppress_packages:
        suppress_common_packages()

    if 'IA_LOGGING_FMT' in os.environ and os.environ['IA_LOGGING_FMT'] != '':
        fmt_string = os.environ['IA_LOGGING_FMT']
    else:
        fmt_string = get_fmt_string(code=code, env=env)

    handlers = list()

    if stdout:
        handlers.append(logging.StreamHandler(sys.stdout))

    if stderr:
        handlers.append(logging.StreamHandler(sys.stderr))

    if logfile:
        handlers.append(logging.FileHandler(logfile))

    if 'IA_LOGFILE' in os.environ and os.environ['IA_LOGFILE'] != '':
        handlers.append(logging.FileHandler(os.environ['IA_LOGFILE']))

    logging.basicConfig(level=level, handlers=handlers, format=fmt_string, style='{', force=True)





import json
import geojson
import zipfile
import io

from pathlib import Path
from typing import Dict, List, Union
from intelinair_utils import AgmriApi

ANNOTATION_URL_TEMPLATE = 'https://api.{}.intelinair.dev/admin/graphql/index'


class AnnotationsApi:

    def __init__(self, agmri_env: str):
        self.api = AgmriApi(agmri_env)

    def list_annotation_types(self, api_limit_size: int = 100):
        offset = 0
        while True:
            query = """
            query { 
              listTagOptions( 
                max: %s
                offset: %s 
                queryCommand: {
                  domain: "Annotation"
                }
              ) { 
                results { 
                  id 
                  name 
                  description 
                  type {
                    id
                    name
                    keyName
                    domainName
                    __typename
                  }
                  color   
                  __typename 
                } 
                totalCount 
                __typename 
              } 
            } 
            """ % (api_limit_size, offset)

            response = self.api.graphql_request('admin', query=query)

            for tag_option in response['data']['listTagOptions']['results']:
                yield tag_option

            if api_limit_size + offset < response['data']['listTagOptions']['totalCount']:
                offset += api_limit_size
            else:
                break

    def create_annotation_project(self, name: str, description: str = "", label_ids: List[int] = None):
        """Create annotation project.

        Args:
            name: name of created project
            description: description for the project
            label_ids: labels used in the project

        Returns: integer project id

        """
        api_key = 'projectCreate'
        payload = """mutation {
          %s(project: {
                        name: \"%s\",
                        description : \"%s\",
                        labelIds: %s
                                }
                        )
                        {
                            id
                            name
                            description
                            labelIds
                            status
                            errors {
                                message
                            }
                        }
                    }""" % (api_key, name,
                            description,
                            str([] if label_ids is None else label_ids))

        response = self._make_request(payload=payload, api_key=api_key)
        return response['data'][api_key]['id']

    def list_annotation_projects(self, api_limit_size=100):
        offset = 0
        while True:
            query = """
            query { 
              listProjects( 
                max: %s 
                offset: %s 
                queryCommand: {}
              ) { 
                results { 
                  id 
                  name 
                  description 
                  labels { 
                    id 
                    name 
                    color 
                    type { 
                      id 
                      domainName 
                      __typename 
                    } 
                    __typename 
                  } 
                  members { 
                    id 
                    user { 
                      id 
                      token 
                      firstname 
                      lastname 
                      __typename 
                    } 
                    __typename 
                  } 
                  status 
                  configs 
                  __typename 
                } 
                totalCount 
                __typename 
              } 
            } 
            """ % (api_limit_size, offset)

            response = self.api.graphql_request('admin', query=query)

            for project in response['data']['listProjects']['results']:
                yield project

            if api_limit_size + offset < response['data']['listProjects']['totalCount']:
                offset += api_limit_size
            else:
                break

    def delete_annotation_project(self, project_id: int):
        query = """
        mutation  {
          projectDelete(id: %s) {
            success
            error
            __typename
          }
        }
        """ % (project_id,)

        response = self.api.graphql_request('admin', query=query)
        assert response['data']['projectDelete']['success'], 'Failed to delete project'

    def create_annotatable_flight(self, project_id: int, flight_id: int):
        """Create annotatable flight out of ordinary flight.

        Args:
            project_id: project in which to create annotatable flight
            flight_id: ordinary flight id

        Returns:
            id of created annotatable flight

        """
        api_key = 'annotatableFlightCreate'
        payload = """mutation {
                  %s(annotatableFlight: {
                                project : {
                                      id : %d
                                    }
                                flight : {
                                     id : %d
                                    }
                                }
                  )
                  {
                    id
                    errors {
                        message
                    }
                  }
                }""" % (api_key, project_id, flight_id)
        response = self._make_request(payload=payload, api_key=api_key)
        return response['data'][api_key]['id']

    def list_annotatable_flights(self, project_id: int, api_limit_size: int = 100):
        offset = 0
        while True:
            query = """
            query { 
              listAnnotatableFlights( 
                max: %s
                offset: %s 
                queryCommand: {
                  projectId: %s
                }
              ) { 
                results { 
                  id 
                  assignee {
                    id
                    user {
                      id
                      username
                      __typename
                    }
                    __typename
                  }
                  status
                  project {
                    id
                    __typename
                  }
                  flight {
                    id
                    code
                    __typename
                  }
                  notes
                  groupNumber   
                  __typename 
                } 
                totalCount 
                __typename 
              } 
            } 
            """ % (api_limit_size, offset, project_id)

            response = self.api.graphql_request('admin', query=query)

            for annotatable_flight in response['data']['listAnnotatableFlights']['results']:
                yield annotatable_flight

            if api_limit_size + offset < response['data']['listAnnotatableFlights']['totalCount']:
                offset += api_limit_size
            else:
                break

    def add_annotations(self, annotatable_flight_id: int, polygons: Dict, labels: List[str]):
        """Add annotation to annotatable flight.

        Args:
            annotatable_flight_id: id of flight to add annotation
            polygons: polygons to add as annotation
            labels: string annotation labels

        Returns:
            response from api call

        """
        api_key = 'annotationCreate'
        payload = """
                mutation {
                  %s(annotation: {
                      flight: {
                          id  : %d
                          },
                      color: \"#FF9A34\"
                      labels: %s
                      geometry: \"%s\"
                  }) {
                id
                labels
                errors {
                    message
                }                
            }
        }""" % (api_key,
                annotatable_flight_id,
                json.dumps(labels) if labels else '[]',
                geojson.dumps(polygons).replace('"', '\\"'))
        response = self._make_request(payload=payload, api_key=api_key)
        return response

    def download_annotations(self, project_id: int, output_path: Union[str, Path]):
        response = self.api.post('/annotations/export', ignore_json=True, params={
            "projectId": project_id,
            "flightCode": None,
            "labelIds": None
        })

        annotations_zipfile = zipfile.ZipFile(io.BytesIO(response.content))
        assert len(annotations_zipfile.filelist) == 1
        annotations = annotations_zipfile.read(annotations_zipfile.filelist[0])
        with open(output_path, 'wb') as fp:
            fp.write(annotations)

    def _make_request(self, payload: str, api_key: str):
        """Make actual call to graphql with provided payload.

        Args:
            payload: request payload
            api_key: name of api in use

        Returns:

        """
        response = self.api.graphql_request(service='admin', query=payload)
        errors = response["errors"] if 'errors' in response else response["data"][api_key].get("errors")
        if errors:
            raise Exception(f'Errors while making api request to {api_key}: {errors}')
        return response



import boto3


def get_ssm_parameter(ssm_client: boto3.client, parameter_name: str) -> str:
    """Returns the value of a SSM parameter.

    Args:
        ssm_client (boto3.client): boto3 ssm client
        parameter_name (str): parameter to get from ssm store

    Returns:
        str: paramenter value
    """
    parameter = ssm_client.get_parameter(Name=parameter_name, WithDecryption=True)
    content = parameter['Parameter']['Value']
    return content



"""Collection of general utility functions"""
import boto3
import os
import logging
import multiprocessing

from typing import Tuple, Union
from pathlib import Path
from botocore.exceptions import ClientError

__all__ = ['split_s3_path', 'download_s3_file', 's3_file_exists', 'upload_s3_file', 'download_from_s3', 'upload_to_s3']

s3_client = boto3.client('s3')
s3_resource = boto3.resource('s3')

logger = logging.getLogger(__name__)


def split_s3_path(s3_path: str) -> Tuple[str, str]:
    """Splits an s3 path into its bucket and key

    Args:
        s3_path: as s3 uri to split ex: 's3://your-bucket/your/key

    Returns:
        bucket: the s3 bucket your object is located in
        key: the key of your object
    """
    assert 's3://' == s3_path[0:5]
    path_parts = s3_path.replace("s3://", "").split("/")
    bucket = path_parts.pop(0)
    key = "/".join(path_parts)

    return bucket, key


def download_s3_file(s3_path: str, output_path: Union[str, Path]) -> None:
    """Downloads the file from s3 to the given path."""
    logger.info(f'Downloading {s3_path} to {output_path}')

    output_path = Path(output_path)
    bucket, key = split_s3_path(s3_path)

    if output_path.is_dir():
        output_path = output_path.joinpath(key.split('/')[-1])

    output_path.parent.mkdir(parents=True, exist_ok=True)
    s3_client.download_file(bucket, key, output_path.as_posix())


def download_from_s3(s3_path: str, local_path: Union[str, Path], num_processes: int = 8) -> None:
    """Downloads either a file or directory from s3"""
    local_path = Path(local_path)

    if s3_file_exists(s3_path) and not s3_path.endswith('/'):  # is actually a file
        return download_s3_file(s3_path, local_path.as_posix())

    def generate_args():
        bucket, prefix = split_s3_path(s3_path)
        bucket = s3_resource.Bucket(bucket)
        for obj in bucket.objects.filter(Prefix=prefix):
            if obj.key[-1] == '/':
                continue
            target = local_path.joinpath(obj.key.replace(prefix, '').lstrip('/'))
            yield f's3://{bucket.name}/{obj.key}', target.as_posix()

    with multiprocessing.Pool(num_processes) as pool:
        pool.starmap(download_s3_file, generate_args())


def upload_s3_file(local_path: str, s3_path: str) -> None:
    """Uploads the files from local_path to s3_path"""
    logger.info(f'Uploading {local_path} to {s3_path}')
    bucket, key = split_s3_path(s3_path)
    s3_client.upload_file(local_path, bucket, key)


def upload_to_s3(local_path: Union[str, Path], s3_path: str, num_processes: int = 8) -> None:
    """Uploads either a file or directory to s3"""
    local_path = Path(local_path)

    if local_path.is_file():
        return upload_s3_file(local_path.as_posix(), s3_path)

    bucket, key = split_s3_path(s3_path)

    def generate_args():
        for root, dirs, files in os.walk(local_path.as_posix()):

            for filename in files:

                # construct the full local path
                local_file = os.path.join(root, filename)

                # construct the full Dropbox path
                relative_path = os.path.relpath(local_file, local_path.as_posix())
                s3_file = os.path.join(key, relative_path)

                # relative_path = os.path.relpath(os.path.join(root, filename))
                yield local_file, f's3://{bucket}/{s3_file}'

    with multiprocessing.Pool(num_processes) as pool:
        pool.starmap(upload_s3_file, generate_args())


def s3_file_exists(s3_path: str) -> bool:
    """Check's if the file exists in a quick way"""
    bucket, key = split_s3_path(s3_path)
    try:
        s3_client.head_object(Bucket=bucket, Key=key)
        return True
    except ClientError:
        return False





import logging
import requests

from datetime import datetime
from typing import Generator, List, Optional
from urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter

from intelinair_utils.api import Api

logger = logging.getLogger(__name__)

KNOWN_ENVS = ['prod', 'release', 'bellflower', 'local']


class MLApi(Api):
    """
    Wrapper class for authenticated queries against the ML Service

    After initialization, authentication has been completed and requests
    through this object will be automatically authenticated.
    """

    def __init__(self, environment):
        # version_string == '' corresponds to the first version of API
        assert environment in KNOWN_ENVS, 'Unknown environment for the ML Service'
        self.environment = environment
        if environment == 'local':
            self.api_url = 'http://localhost:8080'
        else:
            self.api_url = f'https://ml.{environment}.int.intelinair.dev'
        # all calls (other than login) should use the version string
        self.session = requests.Session()
        retries = Retry(total=10, read=5, connect=5, backoff_factor=1, allowed_methods=False,
                        status_forcelist=[104, 500, 502, 503, 504])
        self.session.mount('https://', HTTPAdapter(max_retries=retries))
        self.session.mount('http://', HTTPAdapter(max_retries=retries))

    def get(self, path, params=None, json=None, ignore_json=False):
        res = self.session.get(self.api_url + path, params=params, json=json)
        res.raise_for_status()
        if ignore_json is True:
            return res
        else:
            return res.json()

    def post(self, path, json=None, params=None):
        res = self.session.post(self.api_url + path, json=json, params=params)
        res.raise_for_status()
        return res.json()

    def patch(self, path, json=None, params=None):
        res = self.session.patch(self.api_url + path, json=json, params=params)
        res.raise_for_status()
        return res.json()

    def create_dataset(self, name: str, version: int, data: dict, tags: dict = None) -> dict:
        """Creates a dataset with the given parameters"""
        return self.post('/datasets/', json={
            'name': name,
            'version': version,
            'data': data,
            'tags': tags
        })

    def update_dataset(self, dataset: dict) -> dict:
        """Updates a dataset object in the service with new data"""
        return self.patch(f'/datasets/{dataset["id"]}', json=dataset)

    def get_dataset_by_id(self, dataset_id: int) -> dict:
        """Gets a dataset by its id"""
        return self.get(f'/datasets/{dataset_id}')

    def get_datasets(self, dataset_name: str = None, dataset_version: int = None, api_limit_size: int = 100) -> Generator[dict, None, None]:
        """Gets a generator of datasets optionally filtering by name and/or version"""
        skip = 0
        while True:
            datasets = self.get(
                '/datasets/',
                params={
                    'skip': skip,
                    'limit': api_limit_size,
                    'name': dataset_name,
                    'version': dataset_version
                }
            )
            for dataset in datasets:
                yield dataset
            if len(datasets) == api_limit_size:
                skip += api_limit_size
            else:
                break

    def get_dataset(self, dataset_name: str, dataset_version: int = None) -> dict:
        """Gets a dataset by name and optionally a specific version of the dataset"""
        datasets = sorted(list(self.get_datasets(dataset_name, dataset_version)), key=lambda x: x['version'])
        if len(datasets) == 0:
            raise Exception(f'Could not find dataset with  '
                            f'dataset_name={dataset_name} and version={dataset_version}')
        return datasets[-1]

    def create_model(self, model_name: str, model_version: int, model_hash: str, s3_location: str,
                     tags: dict = None, data: dict = None) -> dict:
        """Creates a model with the ML Service with the given parameters"""
        return self.post(
            '/models/',
            json={
                "model_name": model_name,
                "model_version": model_version,
                "model_hash": model_hash,
                "s3_location": s3_location,
                "tags": tags,
                "data": data
            }
        )

    def update_model(self, model: dict) -> dict:
        """Updates a model"""
        return self.patch(f'/models/{model["id"]}', json=model)

    def get_or_create_model_from_script(self, git_repo_name: str, script_name: str, git_description: str) -> dict:
        """Gets a model by script name and git description or creates one if it doesn't exist"""
        return self.post('/models/script', params={
            'git_repo_name': git_repo_name,
            'script_name': script_name,
            'git_description': git_description
        })

    def get_model_by_id(self, model_id: int) -> dict:
        """Gets a model by id"""
        return self.get(f'/models/{model_id}')

    def get_latest_model(self, model_name: str) -> dict:
        """Gets a model by id"""
        return self.get(f'/models/latest', params={'model_name': model_name})

    def get_models(self, model_name: str = None, model_version: int = None, api_limit_size: int = 100) -> Generator[dict, None, None]:
        """Gets models from the api with optional filters"""
        skip = 0
        while True:
            models = self.get(f'/models/?skip={skip}&limit={api_limit_size}',
                              json={'model_name': model_name, 'model_version': model_version})
            for model in models:
                yield model
            if len(models) == api_limit_size:
                skip += api_limit_size
            else:
                break

    def get_model(self, model_name: str, model_version: int = None) -> dict:
        """Gets the model with the give name and version, if no version is specified the latest version is returned"""
        models = sorted(list(self.get_models(model_name, model_version)), key=lambda x: x['model_version'])
        if len(models) == 0:
            raise Exception(f'Could not find registered models with '
                            f'model_name={model_name} and version={model_version}')
        return models[-1]

    def create_inference(self, flight_code: str, inference_type: str, model_id: int, docker_image: str,
                         hparams: dict, result: dict) -> dict:
        """Creates an inference with the given parameters"""
        return self.post('/inferences/flight/', json={
            'flight_code': flight_code,
            'inference_type': inference_type,
            'model_id': model_id,
            'docker_image': docker_image,
            'hparams': hparams,
            'result': result
        })

    def create_inference_from_script(self, flight_code: str, inference_type: str, result: dict, git_repo_name: str,
                                     script_name: str, git_description: str, hparams: dict,
                                     docker_image: Optional[str] = None):
        """Creates an inference from a script"""
        return self.post('/inferences/flight/script', json={
            'flight_code': flight_code,
            'inference_type': inference_type,
            'git_repo_name': git_repo_name,
            'script_name': script_name,
            'git_description': git_description,
            'docker_image': docker_image,
            'hparams': hparams,
            'result': result
        })

    def update_inference(self, inference: dict) -> dict:
        return self.patch(f'/inferences/flight/{inference["id"]}', json=inference)

    def get_inference_by_id(self, inference_id: int) -> dict:
        """Gets an inference by id"""
        return self.get(f'/inferences/flight/{inference_id}')

    def get_inferences(self, flight_code: List[str] = None, inference_type: str = None, model_id: int = None,
                       model_name: str = None, model_version: int = None, git_repo_name: str = None,
                       script_name: str = None, git_description: str = None, docker_image: str = None,
                       hparams: dict = None, start_ts: datetime = None, end_ts: datetime = None,
                       api_limit_size: int = 100) -> Generator[dict, None, None]:
        """Returns inferences that match the provided filters"""
        request = {
            'flight_code': flight_code,
            'inference_type': inference_type,
            'start_ts': str(start_ts) if start_ts else None,
            'end_ts': str(end_ts) if end_ts else None,
            'model_id': model_id,
            'model_name': model_name,
            'model_version': model_version,
            'git_repo_name': git_repo_name,
            'script_name': script_name,
            'git_description': git_description,
            'docker_image': docker_image,
            'hparams': hparams
        }

        skip = 0
        while True:
            inferences = self.get(f'/inferences/flight/?skip={skip}&limit={api_limit_size}', json=request)
            for inference in inferences:
                yield inference
            if len(inferences) == api_limit_size:
                skip += api_limit_size
            else:
                break

    def get_inference_types(self, flight_code: str) -> List[str]:
        """Returns the inference types available for this flight"""
        return self.get(f'/inferences/flight/types/{flight_code}')

    def get_flight_codes_with_inference_type(self, inference_type: str,
                                             api_limit_size: int = 10000) -> Generator[str, None, None]:
        """Returns a list of flight codes with an inference with the given inference type"""
        skip = 0
        while True:
            flight_codes = self.get(f'/inferences/flight/codes/{inference_type}?skip={skip}&limit={api_limit_size}')
            for flight_code in flight_codes:
                yield flight_code
            if len(flight_codes) == api_limit_size:
                skip += api_limit_size
            else:
                break

    def create_flight_evaluation(self, flight_code: str, inference_id: int, scores: dict) -> dict:
        """Creates an evaluation for a flight with the given parameters"""
        return self.post('/evaluations/flight/', json={
            'flight_code': flight_code,
            'inference_id': inference_id,
            'scores': scores,
        })

    def get_flight_evaluation_by_id(self, flight_evaluation_id: int) -> dict:
        """Returns an evaluation by id"""
        return self.get(f'/evaluations/flight/{flight_evaluation_id}')

    def get_flight_evaluations(self, flight_code: List[str] = None, inference_type: str = None,
                               model_id: int = None, model_name: str = None, model_version: int = None,
                               git_repo_name: str = None, script_name: str = None, git_description: str = None,
                               docker_image: str = None, hparams: dict = None,
                               inference_id: int = None, api_limit_size: int = 100) -> Generator[dict, None, None]:
        """Generates a set of evaluations that match the provided filters"""
        skip = 0
        while True:
            evaluations = self.get(
                '/evaluations/flight/',
                json={
                    'flight_code': flight_code,
                    'inference_type': inference_type,
                    'model_id': model_id,
                    'model_name': model_name,
                    'model_version': model_version,
                    'git_repo_name': git_repo_name,
                    'script_name': script_name,
                    'git_description': git_description,
                    'docker_image': docker_image,
                    'hparams': hparams,
                    'inference_id': inference_id,
                },
                params={
                    'skip': skip,
                    'limit': api_limit_size,
                }
            )
            for evaluation in evaluations:
                yield evaluation
            if len(evaluations) == api_limit_size:
                skip += api_limit_size
            else:
                break

    def create_dataset_evaluation(self, dataset_id: int, model_id: int, scores: dict, hparams: dict, tags: dict = None,
                                  docker_image: str = None) -> dict:
        """Creates an evaluation for the given dataset and model"""
        return self.post('/evaluations/dataset/', json={
            'dataset_id': dataset_id,
            'model_id': model_id,
            'scores': scores,
            'hparams': hparams,
            'tags': tags,
            'docker_image': docker_image
        })

    def get_dataset_evaluation_by_id(self, dataset_evaluation_id: int) -> dict:
        """Returns a dataset evaluation by id"""
        return self.get(f'/evaluations/dataset/{dataset_evaluation_id}')

    def get_dataset_evaluations(self, model_id: id = None, docker_image: str = None, hparams: dict = None,
                                dataset_id: int = None, api_limit_size: int = 100) -> Generator[dict, None, None]:
        """Generates a set of dataset evaluations that match the given parameters"""
        skip = 0
        while True:
            evaluations = self.get(
                '/evaluations/dataset/',
                json={
                    'model_id': model_id,
                    'docker_image': docker_image,
                    'hparams': hparams,
                    'dataset_id': dataset_id,
                },
                params={
                    'skip': skip,
                    'limit': api_limit_size,
                }
            )
            for evaluation in evaluations:
                yield evaluation
            if len(evaluations) == api_limit_size:
                skip += api_limit_size
            else:
                break




import setuptools

with open('src/intelinair_utils/__version__.py') as version_file:
    exec(version_file.read())

with open("requirements.txt", "r") as requirements_file:
    requirements = requirements_file.read().splitlines()

required = []

# Do not add to required lines pointing to Git repositories
EGG_MARK = '#egg='
for line in requirements:
    if line.startswith('-e git:') or line.startswith('-e git+') or line.startswith('git:') or line.startswith('git+'):
        if EGG_MARK in line:
            package_name = line[line.find(EGG_MARK) + len(EGG_MARK):]
            required.append(f'{package_name} @ {line}')
        else:
            print('Dependency to a git repository should have the format:')
            print('git+ssh://git@github.com/xxxxx/xxxxxx#egg=package_name')
    else:
        required.append(line)

setuptools.setup(
    name='intelinair_utils',
    version=globals()['__version__'],
    package_dir={'': 'src'},
    packages=setuptools.find_namespace_packages(where='src'),
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    install_requires=required
)





from typing import Tuple, Union
import logging
import subprocess
import os
import json
from intelinair_utils.pretty_popen import PrettyPopen

__all__ = ["get_error_message", "call_command"]

logger = logging.getLogger(__name__)


def get_error_message(output_folder: str, return_code: int) -> Union[str, None]:
    """Reads the error.json or error.txt file from the
    `output_folder` and returns the content

    Args:
        output_folder (str): folder where is located error.[json, txt] file
        return_code (int): process return code

    Returns:
        Union[str, None]: None if file doesn't exist or file content
    """
    filename = os.path.join(output_folder, 'error.json')
    if os.path.exists(filename):
        try:
            with open(filename) as FL:
                error_data = json.load(FL)
                if "error_code" in error_data and "message" in error_data:
                    return error_data
        except Exception:
            if return_code != 0:
                logger.error("Exception caught in get_error_message!")
                logger.exception("Error while reading error.json")
    else:
        filename = os.path.join(output_folder, 'error.txt')
        try:
            if not os.path.exists(filename):
                return None
            with open(filename) as f:
                error_code = f.readline()[:-1]
                error_message = f.readline().strip()
                return error_code + ": " + error_message
        except Exception:
            if return_code != 0:
                logger.error('Exception caught in get_error_message!')
                logger.exception("Error while reading error.txt")
    return


def call_command(cmd: str, command_name: str, output_folder: str = None,
                 silent: bool = False, print_command: bool = True) -> Tuple[int, str, str]:
    """Calls a shell command, prints stdout/stderr and returns the return code
    Known issue: stdout and stderr are not synced
    Also tries to read error.txt from output_folder and return it
    Finally, stderr is also returned

    Args:
        cmd (str): command (like in Unix shell)
        command_name (str): name of the command (for logging)
        output_folder (str, optional): The output folder where the command 
                                       will save error.[json,txt] outpur. Defaults to None.
        silent (bool, optional): Prints the command output on failure. Defaults to False.
        print_command (bool, optional): if true will print the command. Defaults to True.

    Returns:
        [Tuple]: exit code of the command, error message from error.txt, stderr output
    """
    if not silent and print_command:
        logger.info("Calling {}".format(cmd))

    # TODO: timeout?
    process = PrettyPopen(args=cmd,
                          shell=True,
                          stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE)

    out, err = process.communicate()

    if out and not silent:
        logger.info(out.decode('utf-8', 'ignore'))
    if err:
        logger.info("Subprocess stderr:")
        logger.info(err.decode('utf-8', 'ignore'))

    if process.returncode != 0 or not silent:
        logger.info("Command {} is completed with error code {}".format(
            command_name, process.returncode))

    if output_folder is not None:
        error_msg_from_txt = get_error_message(output_folder, process.returncode)
        return process.returncode, error_msg_from_txt, err.decode('utf-8', 'ignore')
    error_message = None
    if process.returncode:
        error_message = f"Return code: {process.returncode}"

    return process.returncode, error_message, err.decode('utf-8', 'ignore')




import json
import os.path
import boto3
import string
import botocore.exceptions
import logging
import filelock

logger = logging.getLogger(__name__)

PROCESSING_TOKENS_TEMPLATE = "/global/processing/tokens/{environment}/{username}"

__all__ = ['SSMStore', 'FileStore']


class SSMStore:

    def __init__(self, environment, username):
        self.environment = environment
        self.username = self.clean_username(username)

    @staticmethod
    def clean_username(username):
        return ''.join([c if c in string.ascii_letters + string.digits else '_' for c in username])

    def get_tokens(self):
        ssm = boto3.client('ssm')
        parameter = ssm.get_parameter(
            Name=PROCESSING_TOKENS_TEMPLATE.format(environment=self.environment, username=self.username),
            WithDecryption=True
        )
        return json.loads(parameter['Parameter']['Value'])

    def store_tokens(self, tokens):
        try:
            ssm = boto3.client('ssm')
            ssm.put_parameter(
                Name=PROCESSING_TOKENS_TEMPLATE.format(environment=self.environment, username=self.username),
                Value=json.dumps(tokens),
                Type='SecureString',
                Overwrite=True
            )
        except botocore.exceptions.ClientError as e:
            # Boto3 classifies all AWS service errors and exceptions as ClientError exceptions
            logger.warning(f"Failed to store token to the parameter store with exception {e.response['Error']['Code']}")


class FileStore:

    def __init__(self, environment, username, path=None):
        self.environment = environment
        self.username = username
        self.token_path = path if path is not None else os.path.expanduser(f'~/.agmri.{environment}.tokens')
        self.token_lock_path = self.token_path + '.lock'

    def get_tokens(self):
        with filelock.FileLock(self.token_lock_path):
            with open(self.token_path, 'r') as fp:
                return json.load(fp)

    def store_tokens(self, tokens):
        with filelock.FileLock(self.token_lock_path):
            with open(self.token_path, 'w') as fp:
                json.dump(tokens, fp)

